@@ -2,13 +2,15 @@
 Copyright 2021 Objectiv B.V.
 """
 import json
+import operator
+from functools import reduce
 from abc import abstractmethod
 from typing import Dict, Union, TYPE_CHECKING, Tuple, cast, Optional, List, Any, TypeVar, Generic
 
 from sqlalchemy.engine import Dialect
 
 from bach.series import Series
-from bach.expression import Expression
+from bach.expression import Expression, join_expressions
 from bach.series.series import WrappedPartition, ToPandasInfo
 from bach.sql_model import BachSqlModel
 from bach.types import DtypeOrAlias, StructuredDtype, AllSupportedLiteralTypes, Dtype
@@ -235,16 +237,10 @@ class SeriesJson(Series):
         return cast('SeriesBoolean', result)  # we told _binary_operation to return dtype='bool'
 
     def __le__(self, other: Union['Series', AllSupportedLiteralTypes]) -> 'SeriesBoolean':
-        if is_postgres(self.engine):
-            return self._comparator_operation(other, "<@")
-        message_override = f'Operator <= is not supported for type json on database {self.engine.name}'
-        raise DatabaseNotSupportedException(self.engine, message_override=message_override)
+        raise NotImplementedError()
 
     def __ge__(self, other: Union['Series', AllSupportedLiteralTypes]) -> 'SeriesBoolean':
-        if is_postgres(self.engine):
-            return self._comparator_operation(other, "@>")
-        message_override = f'Operator >= is not supported for type json on database {self.engine.name}'
-        raise DatabaseNotSupportedException(self.engine, message_override=message_override)
+        raise NotImplementedError()
 
     def min(self, partition: WrappedPartition = None, skipna: bool = True):
         """ INTERNAL: Only here to not trigger errors from describe """
@@ -406,6 +402,18 @@ class JsonAccessor(Generic[TSeriesJson]):
         # So for now we have a dedicated len function for arrays.
         return self._implementation.get_array_length()
 
+    def array_contains(self, item: Union[int, float, bool, str, None]) -> 'SeriesBoolean':
+        """
+        Find if item is contained in the array.
+
+        :param item: item to be verified if it is contained by the array. Only supports scalar values.
+        :returns: boolean series indicating if the array contains the element.
+
+        This assumes the top-level item in the json is an array.
+        """
+
+        return self._implementation.array_contains(item)
+
 
 class JsonBigQueryAccessorImpl(Generic[TSeriesJson]):
     """
@@ -437,7 +445,11 @@ class JsonBigQueryAccessorImpl(Generic[TSeriesJson]):
         return self._series_object\
             .copy_override(expression=json_str_expression)
 
-    def _get_slice_partial_expr(self, value: Optional[int], is_start: bool) -> Expression:
+    def _get_slice_partial_expr(
+        self,
+        value: Optional[Union[Dict[str, str], int]],
+        is_start: bool,
+    ) -> Expression:
         """
         Return expression for either the lower bound or upper bound of a slice.
 
@@ -448,6 +460,9 @@ class JsonBigQueryAccessorImpl(Generic[TSeriesJson]):
         :return: Expression that will evaluate to an integer that can be used to compare against positions
                     in the array.
         """
+        if isinstance(value, dict):
+            return self._find_in_json_list(value, is_start)
+
         if value is not None and not isinstance(value, int):
             raise TypeError(f'Slice value must be None or an integer, value: {value}')
 
@@ -465,16 +480,54 @@ class JsonBigQueryAccessorImpl(Generic[TSeriesJson]):
             array_len = self.get_array_length()
             return Expression.construct(f'({{}} {value})', array_len)
 
+    def _find_in_json_list(self, filtering_slice: Dict[str, str], is_start: bool):
+        """
+        Return expression for the position of the element that contains the filtering values
+
+        Assumes that self._series_object is an array and each element is a json!
+
+        :param filtering_slice: dictionary containing the keys and values to use on filter.
+        :param is_start: whether value is the slice.start (True) or slice.stop (False) value. If true,
+            the first element containing the values is returned, otherwise the last.
+        :return: Expression that gets the position of the element based on the slicing filters.
+        """
+        if not isinstance(filtering_slice, dict):
+            raise TypeError(f'key should be a dict, actual type: {type(filtering_slice)}')
+
+        # this way we can reuse code from the accessor without duplicating.
+        # 'element' will be filled by iterating over the json array in the query below.
+        element_json = self._series_object.copy_override(
+            expression=Expression.construct('element'),
+        )
+
+        all_filters = [
+            element_json.json.get_value(filter_key, as_str=True) == str(filter_value)
+            for filter_key, filter_value in filtering_slice.items()
+        ]
+        slicing_mask = reduce(operator.and_, all_filters)
+        fmt = (
+            # end slicing is inclusive
+            f"(select {'min' if is_start else '1 + max'}(case when {{}} then pos end) "
+            f"from unnest(JSON_QUERY_ARRAY({{}}, '$')) element with offset as pos)"
+        )
+        return Expression.construct(fmt, slicing_mask, self._series_object)
+
     def get_array_item(self, key: int) -> 'TSeriesJson':
-        """ For documentation, see implementation in parent class :class:`JsonAccessor` """
-        if key >= 0:
+        """
+        Returns an item from the json array.
+        The key is treated as a 0-based index. If negative this will count from the end of the array (one
+            based). If the index does not exist this will render None/NULL.
+        This assumes the top-level item in the json is an array
+        """
+        if key < 0:
+            # BigQuery doesn't (yet) natively support this, so we emulate this by reversing the array
+            key = abs(key)
+            expression = Expression.construct(
+                f'ARRAY_REVERSE(JSON_QUERY_ARRAY({{}}))[SAFE_ORDINAL({key})]',
+                self._series_object
+            )
+        else:
             expression = Expression.construct(f'''JSON_QUERY({{}}, '$[{key}]')''', self._series_object)
-            return self._series_object.copy_override(expression=expression)
-        # case key < 0
-        # BigQuery doesn't (yet) natively support this, so we emulate this.
-        array_len = self.get_array_length()
-        expr_offset = Expression.construct(f'OFFSET({{}} {key})', array_len)
-        expression = Expression.construct('JSON_QUERY_ARRAY({})[{}]', self._series_object, expr_offset)
         return self._series_object.copy_override(expression=expression)
 
     def get_dict_item(self, key: str) -> 'TSeriesJson':
@@ -482,8 +535,7 @@ class JsonBigQueryAccessorImpl(Generic[TSeriesJson]):
         return cast('TSeriesJson', self.get_value(key=key, as_str=False))
 
     def get_value(self, key: str, as_str: bool = False) -> Union['SeriesString', 'TSeriesJson']:
-        """ For documentation, see implementation in parent class :class:`JsonAccessor` """
-        # TODO: as_str is never used?
+        """ For documentation, see function :meth:`JsonAccessor.get_value()` """
         # Special characters are tricky. According to the latests jsonpath spec draft we should use the
         # index selector [1] (e.g. `$['key with special characters']`). But BigQuery only supports the dot
         # selector [2][3]. However BigQuery does allow us to quote the key when using the dot selector [4],
@@ -505,16 +557,20 @@ class JsonBigQueryAccessorImpl(Generic[TSeriesJson]):
         if not as_str:
             return self._series_object.copy_override(expression=json_expression)
         from bach.series import SeriesString
+        # JSON_QUERY will return a string containing valid json. i.e. if the queried value is a string, then
+        # it will be quoted. Here we remove the quotes
         value_expression = Expression.construct(
-            '''JSON_VALUE({}, '$."{}"')''',
-            self._series_object,
-            Expression.raw(key)
+            f'REGEXP_REPLACE({{}}, {{}}, {{}})',
+            json_expression,
+            Expression.raw('r\'(^"|"$)\''),
+            Expression.string_value(''),
         )
-        expression = Expression.construct('COALESCE({}, {})', value_expression, json_expression)
-        return self._series_object.copy_override(expression=expression).copy_override_type(SeriesString)
+        return self._series_object.copy_override(
+            expression=value_expression
+        ).copy_override_type(SeriesString)
 
     def get_array_length(self) -> 'SeriesInt64':
-        """ For documentation, see implementation in parent class :class:`JsonAccessor` """
+        """ For documentation, see function :meth:`JsonAccessor.get_array_length()` """
         # Implementing a more generic __len__() function is not trivial as a json object can be (among
         # others) an array, a dict, or a string, all of which should be supported by a generic __len__().
         # So for now we have a dedicated len function for arrays.
@@ -524,6 +580,20 @@ class JsonBigQueryAccessorImpl(Generic[TSeriesJson]):
             .copy_override_type(SeriesInt64) \
             .copy_override(expression=expression)
 
+    def array_contains(self, item: Union[int, float, bool, str, None]) -> 'SeriesBoolean':
+        """ For documentation, see implementation in class :class:`JsonAccessor` """
+        item_json = json.dumps(item)
+        in_expr = Expression.construct(
+            '(SELECT {} IN UNNEST(JSON_QUERY_ARRAY({})))',
+            Expression.string_value(item_json),
+            self._series_object
+        )
+        expression = Expression.construct('IF({} IS NULL, NULL, {})', self._series_object, in_expr)
+        from bach.series import SeriesBoolean
+        return self._series_object\
+            .copy_override(expression=expression)\
+            .copy_override_type(SeriesBoolean)
+
 
 class JsonPostgresAccessorImpl(Generic[TSeriesJson]):
     """
@@ -596,7 +666,12 @@ class JsonPostgresAccessorImpl(Generic[TSeriesJson]):
             )
 
     def get_array_item(self, key: int) -> 'TSeriesJson':
-        """ For documentation, see implementation in parent class :class:`JsonAccessor` """
+        """
+        Returns an item from the json array.
+        The key is treated as a 0-based index. If negative this will count from the end of the array (one
+            based). If the index does not exist this will render None/NULL.
+        This assumes the top-level item in the json is an array
+        """
         return self._series_object \
             .copy_override(expression=Expression.construct(f'{{}}->{key}', self._series_object))
 
@@ -605,7 +680,7 @@ class JsonPostgresAccessorImpl(Generic[TSeriesJson]):
         return cast('TSeriesJson', self.get_value(key=key, as_str=False))
 
     def get_value(self, key: str, as_str: bool = False) -> Union['SeriesString', 'TSeriesJson']:
-        """ For documentation, see implementation in parent class :class:`JsonAccessor` """
+        """ For documentation, see function :meth:`JsonAccessor.get_value()` """
 
         if '"' in key:
             raise ValueError(f'key values containing double quotes are not supported. key: {key}')
@@ -623,9 +698,13 @@ class JsonPostgresAccessorImpl(Generic[TSeriesJson]):
         return result
 
     def get_array_length(self) -> 'SeriesInt64':
-        """ For documentation, see implementation in parent class :class:`JsonAccessor` """
+        """ For documentation, see function :meth:`JsonAccessor.get_array_length()` """
         from bach.series import SeriesInt64
         expression = Expression.construct('jsonb_array_length({})', self._series_object)
         return self._series_object \
             .copy_override_type(SeriesInt64) \
             .copy_override(expression=expression)
+
+    def array_contains(self, item: Union[int, float, bool, str, None]) -> 'SeriesBoolean':
+        """ For documentation, see implementation in class :class:`JsonAccessor` """
+        return self._series_object._comparator_operation([item], "@>")
